@@ -5,6 +5,7 @@ import logging
 import cairo
 import os
 
+from blueman.bluez.Adapter import Adapter
 from blueman.bluez.Battery import Battery
 from blueman.bluez.Device import Device
 from blueman.bluez.Manager import Manager
@@ -16,13 +17,14 @@ from blueman.Constants import PIXMAP_PATH
 from blueman.Functions import launch
 from blueman.Sdp import ServiceUUID, OBEX_OBJPUSH_SVCLASS_ID
 from blueman.gui.GtkAnimation import TreeRowFade, CellFade, AnimBase
-from blueman.main.Config import Config
 from _blueman import ConnInfoReadError, conn_info
 
 import gi
 gi.require_version("Gtk", "3.0")
 from gi.repository import Gtk
 from gi.repository import GLib
+from gi.repository import GObject
+from gi.repository import Gio
 from gi.repository import Gdk
 from gi.repository import GdkPixbuf
 from gi.repository import Pango
@@ -31,13 +33,21 @@ if TYPE_CHECKING:
     from blueman.main.Manager import Blueman
 
 
+class SurfaceObject(GObject.Object):
+    __gtype_name__ = "SurfaceObject"
+
+    def __init__(self, surface: cairo.ImageSurface) -> None:
+        super().__init__()
+        self.surface = surface
+
+
 class ManagerDeviceList(DeviceList):
-    def __init__(self, adapter: Optional[str] = None, inst: Optional["Blueman"] = None) -> None:
+    def __init__(self, inst: "Blueman", adapter: Optional[str] = None) -> None:
         cr = Gtk.CellRendererText()
         cr.props.ellipsize = Pango.EllipsizeMode.END
         tabledata: List[ListDataDict] = [
             # device picture
-            {"id": "device_surface", "type": str, "renderer": Gtk.CellRendererPixbuf(),
+            {"id": "device_surface", "type": SurfaceObject, "renderer": Gtk.CellRendererPixbuf(),
              "render_attrs": {}, "celldata_func": (self._set_cell_data, None)},
             # device caption
             {"id": "caption", "type": str, "renderer": cr,
@@ -48,9 +58,6 @@ class ManagerDeviceList(DeviceList):
             {"id": "rssi_pb", "type": GdkPixbuf.Pixbuf, "renderer": Gtk.CellRendererPixbuf(),
              "render_attrs": {}, "view_props": {"spacing": 0},
              "celldata_func": (self._set_cell_data, "rssi")},
-            {"id": "lq_pb", "type": GdkPixbuf.Pixbuf, "renderer": Gtk.CellRendererPixbuf(),
-             "render_attrs": {}, "view_props": {"spacing": 0},
-             "celldata_func": (self._set_cell_data, "lq")},
             {"id": "tpl_pb", "type": GdkPixbuf.Pixbuf, "renderer": Gtk.CellRendererPixbuf(),
              "render_attrs": {}, "view_props": {"spacing": 0},
              "celldata_func": (self._set_cell_data, "tpl")},
@@ -61,9 +68,7 @@ class ManagerDeviceList(DeviceList):
             {"id": "objpush", "type": bool},  # used to set Send File button
             {"id": "battery", "type": float},
             {"id": "rssi", "type": float},
-            {"id": "lq", "type": float},
             {"id": "tpl", "type": float},
-            {"id": "icon_info", "type": Gtk.IconInfo},
             {"id": "cell_fader", "type": CellFade},
             {"id": "row_fader", "type": TreeRowFade},
             {"id": "initial_anim", "type": bool},
@@ -81,7 +86,7 @@ class ManagerDeviceList(DeviceList):
         self.manager.connect_signal("battery-removed", self.on_battery_removed)
         self._batteries: Dict[str, Battery] = {}
 
-        self.Config = Config("org.blueman.general")
+        self.Config = Gio.Settings(schema_id="org.blueman.general")
         self.Config.connect('changed', self._on_settings_changed)
         # Set the correct sorting
         self._on_settings_changed(self.Config, "sort-by")
@@ -92,8 +97,9 @@ class ManagerDeviceList(DeviceList):
         self.tooltip_col: Optional[Gtk.TreeViewColumn] = None
 
         self.connect("popup-menu", self._on_popup_menu)
-        self.connect("button_press_event", self.on_event_clicked)
-        self.connect("button_release_event", self.on_event_clicked)
+        self.connect("button-press-event", self._on_event_clicked)
+        self.connect("button-release-event", self._on_event_clicked)
+        self.connect("key-press-event", self._on_key_pressed)
 
         self.menu: Optional[ManagerDeviceMenu] = None
 
@@ -106,7 +112,7 @@ class ManagerDeviceList(DeviceList):
         self.set_search_equal_func(self.search_func)
         self.filter.set_visible_func(self.filter_func)
 
-    def _on_settings_changed(self, settings: Config, key: str) -> None:
+    def _on_settings_changed(self, settings: Gio.Settings, key: str) -> None:
         if key in ('sort-by', 'sort-order'):
             sort_by = settings['sort-by']
             sort_order = settings['sort-order']
@@ -145,9 +151,12 @@ class ManagerDeviceList(DeviceList):
         return True
 
     def filter_func(self, _model: Gtk.TreeModel, tree_iter: Gtk.TreeIter, _data: Any) -> bool:
-        no_name = self.get(tree_iter, "no_name")["no_name"]
-        if no_name and self.Config["hide-unnamed"]:
-            logging.debug("Hiding unnamed device")
+        row = self.get(tree_iter, "no_name", "device")
+        device = row["device"]
+        klass = get_minor_class(device["Class"]) if device is not None else None
+
+        if row["no_name"] and self.Config["hide-unnamed"] and klass not in (_("Keyboard"), _("Combo")):
+            logging.info("Hiding unnamed device")
             return False
         else:
             return True
@@ -197,9 +206,6 @@ class ManagerDeviceList(DeviceList):
             return False
 
     def _on_popup_menu(self, _widget: Gtk.Widget) -> bool:
-        if self.Blueman is None:
-            return False
-
         if self.menu is None:
             self.menu = ManagerDeviceMenu(self.Blueman)
 
@@ -212,7 +218,7 @@ class ManagerDeviceList(DeviceList):
 
         return True
 
-    def on_event_clicked(self, _widget: Gtk.Widget, event: Gdk.Event) -> bool:
+    def _on_event_clicked(self, _widget: Gtk.Widget, event: Gdk.Event) -> bool:
         if event.type not in (Gdk.EventType._2BUTTON_PRESS, Gdk.EventType.BUTTON_PRESS):
             return False
 
@@ -223,15 +229,14 @@ class ManagerDeviceList(DeviceList):
             path = posdata[0]
             assert path is not None
 
-        childpath = self.filter.convert_path_to_child_path(path)
-        assert childpath is not None
-        row = self.get(childpath, "device", "connected")
+        tree_iter = self.filter.get_iter(path)
+        assert tree_iter is not None
+        child_iter = self.filter.convert_iter_to_child_iter(tree_iter)
+        assert child_iter is not None
+        row = self.get(child_iter, "device", "connected")
         if not row:
             return False
 
-        if self.Blueman is None:
-            return False
-
         if self.menu is None:
             self.menu = ManagerDeviceMenu(self.Blueman)
 
@@ -239,7 +244,7 @@ class ManagerDeviceList(DeviceList):
             if self.menu.show_generic_connect_calc(row["device"]['UUIDs']):
                 if row["connected"]:
                     self.menu.disconnect_service(row["device"])
-                else:
+                elif Adapter(obj_path=row["device"]["Adapter"])["Powered"]:
                     self.menu.connect_service(row["device"])
 
         if event.type == Gdk.EventType.BUTTON_PRESS and cast(Gdk.EventButton, event).button == 3:
@@ -247,37 +252,46 @@ class ManagerDeviceList(DeviceList):
 
         return False
 
-    def get_icon_info(self, icon_name: str, size: int = 48, fallback: bool = True) -> Optional[Gtk.IconInfo]:
-        if icon_name is None and not fallback:
-            return None
-        elif icon_name is None and fallback:
-            icon_name = "image-missing"
+    def _on_key_pressed(self, _widget: Gtk.Widget, event: Gdk.EventKey) -> bool:
+        if not (event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == Gdk.KEY_c):
+            return False
 
-        icon_info = self.icon_theme.lookup_icon_for_scale(icon_name, size, self.get_scale_factor(),
-                                                          Gtk.IconLookupFlags.FORCE_SIZE)
+        selected = self.selected()
+        if not selected:
+            return False
 
-        return icon_info
+        row = self.get(selected, "device")
+        if not row:
+            return False
 
-    def make_device_icon(self, icon_info: Gtk.IconInfo, is_paired: bool = False, is_trusted: bool = False,
-                         is_blocked: bool = False) -> cairo.Surface:
+        Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).set_text(row["device"]["Address"], -1)
+        return True
+
+    def _load_surface(self, icon_name: str, size: int) -> cairo.ImageSurface:
         window = self.get_window()
         scale = self.get_scale_factor()
-        target = icon_info.load_surface(window)
+        surface = self.icon_theme.load_surface(icon_name, size, scale, window, Gtk.IconLookupFlags.FORCE_SIZE)
+        if surface is None:
+            surface = self.icon_theme.load_surface("image-missing", size, scale, window, Gtk.IconLookupFlags.FORCE_SIZE)
+        assert surface is not None
+
+        return cast(cairo.ImageSurface, surface)
+
+    def _make_device_icon(self, icon_name: str, is_paired: bool, is_connected: bool, is_trusted: bool,
+                          is_blocked: bool) -> cairo.ImageSurface:
+        scale = self.get_scale_factor()
+        target = self._load_surface(icon_name, 48)
         ctx = cairo.Context(target)
 
-        if is_paired:
-            _icon_info = self.get_icon_info("blueman-paired-emblem", 16, False)
-            assert _icon_info is not None
-            paired_surface = _icon_info.load_surface(window)
+        if is_connected or is_paired:
+            icon = "blueman-connected-emblem" if is_connected else "blueman-paired-emblem"
+            paired_surface = self._load_surface(icon, 16)
             ctx.set_source_surface(paired_surface, 1 / scale, 1 / scale)
             ctx.paint_with_alpha(0.8)
 
         if is_trusted:
-            _icon_info = self.get_icon_info("blueman-trusted-emblem", 16, False)
-            assert _icon_info is not None
-            trusted_surface = _icon_info.load_surface(window)
+            trusted_surface = self._load_surface("blueman-trusted-emblem", 16)
             assert isinstance(target, cairo.ImageSurface)
-            assert isinstance(trusted_surface, cairo.ImageSurface)
             height = target.get_height()
             mini_height = trusted_surface.get_height()
             y = height / scale - mini_height / scale - 1 / scale
@@ -286,11 +300,8 @@ class ManagerDeviceList(DeviceList):
             ctx.paint_with_alpha(0.8)
 
         if is_blocked:
-            _icon_info = self.get_icon_info("blueman-blocked-emblem", 16, False)
-            assert _icon_info is not None
-            blocked_surface = _icon_info.load_surface(window)
+            blocked_surface = self._load_surface("blueman-blocked-emblem", 16)
             assert isinstance(target, cairo.ImageSurface)
-            assert isinstance(blocked_surface, cairo.ImageSurface)
             width = target.get_width()
             mini_width = blocked_surface.get_width()
             ctx.set_source_surface(blocked_surface, (width - mini_width - 1) / scale, 1 / scale)
@@ -298,28 +309,31 @@ class ManagerDeviceList(DeviceList):
 
         return target
 
-    def device_remove_event(self, device: Device) -> None:
-        tree_iter = self.find_device(device)
+    def device_remove_event(self, object_path: str) -> None:
+        tree_iter = self.find_device_by_path(object_path)
         assert tree_iter is not None
 
         iter_set, _child_tree_iter = self.filter.convert_child_iter_to_iter(tree_iter)
         if iter_set:
             row_fader = self.get(tree_iter, "row_fader")["row_fader"]
-            self._prepare_fader(row_fader, lambda: self.__fader_finished(device))
+            self._prepare_fader(row_fader, lambda: self.__fader_finished(object_path))
             row_fader.animate(start=row_fader.get_state(), end=0.0, duration=400)
 
-    def __fader_finished(self, device: Device) -> None:
-        super().device_remove_event(device)
-        self.emit("device-selected", None, None)
-
-    def device_add_event(self, device: Device) -> None:
-        self.add_device(device)
+    def __fader_finished(self, object_path: str) -> None:
+        super().device_remove_event(object_path)
 
     @staticmethod
     def make_caption(name: str, klass: str, address: str) -> str:
         return "<span size='x-large'>%(0)s</span>\n<span size='small'>%(1)s</span>\n<i>%(2)s</i>" \
                % {"0": html.escape(name), "1": klass, "2": address}
 
+    @staticmethod
+    def make_display_name(alias: str, klass: int, address: str) -> str:
+        if alias.replace("-", ":") == address:
+            return _("Unnamed device")
+        else:
+            return alias
+
     @staticmethod
     def get_device_class(device: Device) -> str:
         klass = get_minor_class(device['Class'])
@@ -359,10 +373,13 @@ class ManagerDeviceList(DeviceList):
         else:
             description = get_major_class(device['Class'])
 
-        icon_info = self.get_icon_info(device["Icon"], 48, False)
-        caption = self.make_caption(device['Alias'], description, device['Address'])
+        surface = self._make_device_icon(device["Icon"], device["Paired"], device["Connected"], device["Trusted"],
+                                         device["Blocked"])
+        surface_object = SurfaceObject(surface)
+        display_name = self.make_display_name(device.display_name, device["Class"], device['Address'])
+        caption = self.make_caption(display_name, description, device['Address'])
 
-        self.set(tree_iter, caption=caption, icon_info=icon_info, alias=device['Alias'], objpush=has_objpush)
+        self.set(tree_iter, caption=caption, alias=display_name, objpush=has_objpush, device_surface=surface_object)
 
         try:
             self.row_update_event(tree_iter, "Trusted", device['Trusted'])
@@ -426,6 +443,13 @@ class ManagerDeviceList(DeviceList):
     def row_update_event(self, tree_iter: Gtk.TreeIter, key: str, value: Any) -> None:
         logging.info(f"{key} {value}")
 
+        device = self.get(tree_iter, "device")["device"]
+
+        if key in ("Blocked", "Connected", "Paired", "Trusted"):
+            surface = self._make_device_icon(device["Icon"], device["Paired"], device["Connected"], device["Trusted"],
+                                             device["Blocked"])
+            self.set(tree_iter, device_surface=SurfaceObject(surface))
+
         if key == "Trusted":
             if value:
                 self.set(tree_iter, trusted=True)
@@ -439,12 +463,11 @@ class ManagerDeviceList(DeviceList):
                 self.set(tree_iter, paired=False)
 
         elif key == "Alias":
-            device = self.get(tree_iter, "device")["device"]
             c = self.make_caption(value, self.get_device_class(device), device['Address'])
-            self.set(tree_iter, caption=c, alias=value)
+            name = self.make_display_name(device.display_name, device["Class"], device["Address"])
+            self.set(tree_iter, caption=c, alias=name)
 
         elif key == "UUIDs":
-            device = self.get(tree_iter, "device")["device"]
             has_objpush = self._has_objpush(device)
             self.set(tree_iter, objpush=has_objpush)
 
@@ -452,7 +475,7 @@ class ManagerDeviceList(DeviceList):
             self.set(tree_iter, connected=value)
 
             if value:
-                self._monitor_power_levels(tree_iter, self.get(tree_iter, "device")["device"])
+                self._monitor_power_levels(tree_iter, device)
             else:
                 self._disable_power_levels(tree_iter)
         elif key == "Name":
@@ -474,22 +497,18 @@ class ManagerDeviceList(DeviceList):
         # cinfo init may fail for bluetooth devices version 4 and up
         # FIXME Workaround is horrible and we should show something better
         if cinfo.failed:
-            bars.update({"rssi": 100.0, "tpl": 100.0, "lq": 100.0})
+            bars.update({"rssi": 100.0, "tpl": 100.0})
         else:
             try:
                 bars["rssi"] = max(50 + float(cinfo.get_rssi()) / 127 * 50, 10)
             except ConnInfoReadError:
                 bars["rssi"] = 50
-            try:
-                bars["lq"] = max(float(cinfo.get_lq()) / 255 * 100, 10)
-            except ConnInfoReadError:
-                bars["lq"] = 10
             try:
                 bars["tpl"] = max(50 + float(cinfo.get_tpl()) / 127 * 50, 10)
             except ConnInfoReadError:
                 bars["tpl"] = 50
 
-        if row["battery"] == row["rssi"] == row["tpl"] == row["lq"] == 0:
+        if row["battery"] == row["rssi"] == row["tpl"] == 0:
             self._prepare_fader(row["cell_fader"]).animate(start=0.0, end=1.0, duration=400)
 
         w = 14 * self.get_scale_factor()
@@ -502,12 +521,12 @@ class ManagerDeviceList(DeviceList):
                 self.set(tree_iter, **{name: perc, f"{name}_pb": icon})
 
     def _disable_power_levels(self, tree_iter: Gtk.TreeIter) -> None:
-        row = self.get(tree_iter, "cell_fader", "battery", "rssi", "lq", "tpl")
-        if row["battery"] == row["rssi"] == row["tpl"] == row["lq"] == 0:
+        row = self.get(tree_iter, "cell_fader", "battery", "rssi", "tpl")
+        if row["battery"] == row["rssi"] == row["tpl"] == 0:
             return
 
-        self.set(tree_iter, battery=0, rssi=0, lq=0, tpl=0)
-        self._prepare_fader(row["cell_fader"], lambda: self.set(tree_iter, battery_pb=None, rssi_pb=None, lq_pb=None,
+        self.set(tree_iter, battery=0, rssi=0, tpl=0)
+        self._prepare_fader(row["cell_fader"], lambda: self.set(tree_iter, battery_pb=None, rssi_pb=None,
                                                                 tpl_pb=None)).animate(start=1.0, end=0.0, duration=400)
 
     def _prepare_fader(self, fader: AnimBase, callback: Optional[Callable[[], None]] = None) -> AnimBase:
@@ -535,16 +554,15 @@ class ManagerDeviceList(DeviceList):
             tree_iter = self.get_iter(path[0])
             assert tree_iter is not None
 
-            row = self.get(tree_iter, "trusted", "paired", "blocked")
-            trusted = row["trusted"]
-            paired = row["paired"]
-            blocked = row["blocked"]
+            row = self.get(tree_iter, "connected", "trusted", "paired", "blocked")
             str_list = []
-            if trusted:
+            if row["connected"]:
+                str_list.append(_("Connected"))
+            if row["trusted"]:
                 str_list.append(_("Trusted"))
-            if paired:
+            if row["paired"]:
                 str_list.append(_("Paired"))
-            if blocked:
+            if row["blocked"]:
                 str_list.append(_("Blocked"))
 
             text = ", ".join(str_list)
@@ -559,7 +577,6 @@ class ManagerDeviceList(DeviceList):
 
         elif path[1] == self.columns["battery_pb"] \
                 or path[1] == self.columns["tpl_pb"] \
-                or path[1] == self.columns["lq_pb"] \
                 or path[1] == self.columns["rssi_pb"]:
             tree_iter = self.get_iter(path[0])
             assert tree_iter is not None
@@ -572,7 +589,6 @@ class ManagerDeviceList(DeviceList):
 
             battery = self.get(tree_iter, "battery")["battery"]
             rssi = self.get(tree_iter, "rssi")["rssi"]
-            lq = self.get(tree_iter, "lq")["lq"]
             tpl = self.get(tree_iter, "tpl")["tpl"]
 
             if battery != 0:
@@ -600,12 +616,6 @@ class ManagerDeviceList(DeviceList):
                     lines.append(_("Received Signal Strength: %(rssi)u%% <i>(%(rssi_state)s)</i>") %
                                  {"rssi": rssi, "rssi_state": rssi_state})
 
-            if lq != 0:
-                if path[1] == self.columns["lq_pb"]:
-                    lines.append(_("<b>Link Quality: %(lq)u%%</b>") % {"lq": lq})
-                else:
-                    lines.append(_("Link Quality: %(lq)u%%") % {"lq": lq})
-
             if tpl != 0:
                 if tpl < 30:
                     tpl_state = _("Low")
@@ -644,9 +654,8 @@ class ManagerDeviceList(DeviceList):
                        tree_iter: Gtk.TreeIter, data: Optional[str]) -> None:
         tree_iter = model.convert_iter_to_child_iter(tree_iter)
         if data is None:
-            row = self.get(tree_iter, "icon_info", "trusted", "paired", "blocked")
-            surface = self.make_device_icon(row["icon_info"], row["paired"], row["trusted"], row["blocked"])
-            cell.set_property("surface", surface)
+            row = self.get(tree_iter, "device_surface")
+            cell.set_property("surface", row["device_surface"].surface)
         else:
             window = self.get_window()
             scale = self.get_scale_factor()
